#include "adu.h"
#include <dirent.h> /* readdir() */

#include "gcc-compat.h"
#include "cmdline.h"
#include "fd.h"
#include "string.h"
#include "error.h"
#include "portable_io.h"

DEFINE_ERRLIST;
int osl_errno;

/** Command line and config file options. */
static struct gengetopt_args_info conf;

enum uid_info_flags {
	/** whether this slot of the hash table is used. */
	UI_FL_SLOT_USED = 1,
	/** whether this uid should be taken into account. */
	UI_FL_ADMISSIBLE = 2,
};

struct user_info {
	uint32_t uid;
	uint32_t flags;
	struct osl_table *table;
	uint64_t files;
	uint64_t bytes;
	uint64_t dirs;
	struct osl_table_description *desc;
};

/**
 * Contains info for each user that owns at least one regular file.
 *
 * Even users that are not taken into account because of the --uid
 * option occupy a slot in this hash table. This allows to find out
 * quicky whether a uid is admissible. And yes, this has to be fast.
 */
static struct user_info *uid_hash_table;

static inline int ui_used(struct user_info *ui)
{
	return ui->flags & UI_FL_SLOT_USED;
}

static inline int ui_admissible(struct user_info *ui)
{
	return ui->flags & UI_FL_ADMISSIBLE;
}

struct uid_range {
	uint32_t low;
	uint32_t high;
};

static struct uid_range *admissible_uids;

static inline int check_uid_arg(const char *arg, uint32_t *uid)
{
	const uint32_t max = ~0U;
	/*
	 * we need an 64-bit int for string -> uid conversion because strtoll()
	 * returns a signed value.
	 */
	int64_t val;
	int ret = para_atoi64(arg, &val);

	if (ret < 0)
		return ret;
	if (val < 0 || val > max)
		return -ERRNO_TO_ERROR(EINVAL);
	*uid = val;
	return 1;
}

static int parse_uid_range(const char *orig_arg, struct uid_range *ur)
{
	int ret;
	char *arg = para_strdup(orig_arg), *p = strchr(arg, '-');

	if (!p || p == arg) { /* -42 or 42 */
		ret = check_uid_arg(p? p + 1 : arg, &ur->high);
		if (ret < 0)
			goto out;
		ur->low = p? 0 : ur->high;
		ret = 1;
		goto out;
	}
	/* 42- or 42-4711 */
	*p = '\0';
	p++;
	ret = check_uid_arg(arg, &ur->low);
	if (ret < 0)
		goto out;
	ur->high = ~0U;
	if (*p) { /* 42-4711 */
		ret = check_uid_arg(p, &ur->high);
		if (ret < 0)
			goto out;
	}
	if (ur->low > ur->high)
		ret = -ERRNO_TO_ERROR(EINVAL);
out:
	if (ret < 0)
		ERROR_LOG("bad uid option: %s\n", orig_arg);
	else
		INFO_LOG("admissible uid range: %u - %u\n", ur->low,
			ur->high);
	free(arg);
	return ret;
}


/** evaluates to 1 if x < y, to -1 if x > y and to 0 if x == y */
#define NUM_COMPARE(x, y) ((int)((x) < (y)) - (int)((x) > (y)))

/**
 * The log function.
 *
 * \param ll Loglevel.
 * \param fml Usual format string.
 *
 * All XXX_LOG() macros use this function.
 */
__printf_2_3 void __log(int ll, const char* fmt,...)
{
	va_list argp;
	FILE *outfd;
	struct tm *tm;
	time_t t1;
	char str[255] = "";

	if (ll < conf.loglevel_arg)
		return;
	outfd = stderr;
	time(&t1);
	tm = localtime(&t1);
	strftime(str, sizeof(str), "%b %d %H:%M:%S", tm);
	fprintf(outfd, "%s ", str);
	va_start(argp, fmt);
	vfprintf(outfd, fmt, argp);
	va_end(argp);
}

/**
 * Compare the size of two directories
 *
 * \param obj1 Pointer to the first object.
 * \param obj2 Pointer to the second object.
 *
 * This function first compares the size values as usual integers. If they compare as
 * equal, the address of \a obj1 and \a obj2 are compared. So this compare function
 * returns zero if and only if \a obj1 and \a obj2 point to the same memory area.
 */
static int size_compare(const struct osl_object *obj1, const struct osl_object *obj2)
{
	uint64_t d1 = *(uint64_t *)obj1->data;
	uint64_t d2 = *(uint64_t *)obj2->data;
	int ret = NUM_COMPARE(d2, d1);

	if (ret)
		return ret;
	//INFO_LOG("addresses: %p, %p\n", obj1->data, obj2->data);
	return NUM_COMPARE(obj2->data, obj1->data);
}

/**
 * Compare two osl objects pointing to unsigned integers of 64 bit size.
 *
 * \param obj1 Pointer to the first integer.
 * \param obj2 Pointer to the second integer.
 *
 * \return The values required for an osl compare function.
 *
 * \sa osl_compare_func, osl_hash_compare().
 */
static int uint64_compare(const struct osl_object *obj1,
		const struct osl_object *obj2)
{
	uint64_t d1 = read_u64((const char *)obj1->data);
	uint64_t d2 = read_u64((const char *)obj2->data);

	if (d1 < d2)
		return 1;
	if (d1 > d2)
		return -1;
	return 0;
}

/** The columns of the directory table. */
enum dir_table_columns {
	/** The name of the directory. */
	DT_NAME,
	/** The dir count number. */
	DT_NUM,
	/** The number of the parent directory. */
	DT_PARENT_NUM,
	/** The number of bytes of all regular files. */
	DT_BYTES,
	/** The number of all regular files. */
	DT_FILES,
	/** Number of columns in this table. */
	NUM_DT_COLUMNS
};

static struct osl_column_description dir_table_cols[] = {
	[DT_NAME] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = 0,
		.name = "dir",
	},
	[DT_NUM] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_FIXED_SIZE | OSL_UNIQUE,
		.name = "num",
		.compare_function = uint64_compare,
		.data_size = sizeof(uint64_t)
	},
	[DT_PARENT_NUM] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_FIXED_SIZE | OSL_UNIQUE,
		.name = "parent_num",
		.compare_function = size_compare,
		.data_size = sizeof(uint64_t)
	},
	[DT_BYTES] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags =  OSL_RBTREE | OSL_FIXED_SIZE,
		.compare_function = size_compare,
		.name = "num_bytes",
		.data_size = sizeof(uint64_t)
	},
	[DT_FILES] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags =  OSL_RBTREE | OSL_FIXED_SIZE,
		.compare_function = size_compare,
		.name = "num_files",
		.data_size = sizeof(uint64_t)
	}
};

static struct osl_table_description dir_table_desc = {
	.name = "dir_table",
	.num_columns = NUM_DT_COLUMNS,
	.flags = 0,
	.column_descriptions = dir_table_cols,
};

/** The columns of the id table. */
enum user_table_columns {
	/** The numer of the directory. */
	UT_DIR_NUM,
	/** The number of bytes of all regular files in this dir owned by this id. */
	UT_BYTES,
	/** The number of files in this dir owned by this id. */
	UT_FILES,
	/** Number of columns in this table. */
	NUM_UT_COLUMNS
};

static struct osl_column_description user_table_cols[] = {
	[UT_DIR_NUM] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_FIXED_SIZE | OSL_UNIQUE,
		.name = "dir_num",
		.compare_function = uint64_compare,
		.data_size = sizeof(uint64_t)
	},
	[UT_BYTES] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_FIXED_SIZE,
		.compare_function = size_compare,
		.name = "num_bytes",
		.data_size = sizeof(uint64_t)
	},
	[UT_FILES] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_FIXED_SIZE,
		.compare_function = size_compare,
		.name = "num_files",
		.data_size = sizeof(uint64_t)
	},
};

static struct osl_table *dir_table;

static int add_directory(char *dirname, uint64_t *dir_num, uint64_t *parent_dir_num,
		uint64_t *dir_size, uint64_t *dir_files)
{
	struct osl_object dir_objects[NUM_DT_COLUMNS];

	INFO_LOG("adding #%llu: %s\n", (long long unsigned)*dir_num, dirname);
	dir_objects[DT_NAME].data = dirname;
	dir_objects[DT_NAME].size = strlen(dirname) + 1;
	dir_objects[DT_NUM].data = dir_num;
	dir_objects[DT_NUM].size = sizeof(*dir_num);
	dir_objects[DT_PARENT_NUM].data = parent_dir_num;
	dir_objects[DT_PARENT_NUM].size = sizeof(*parent_dir_num);
	dir_objects[DT_BYTES].data = dir_size;
	dir_objects[DT_BYTES].size = sizeof(*dir_size);
	dir_objects[DT_FILES].data = dir_files;
	dir_objects[DT_FILES].size = sizeof(*dir_files);
	return osl(osl_add_row(dir_table, dir_objects));
}

static uint32_t num_uids;

static int open_user_table(struct user_info *ui, int create)
{
	int ret;

	ui->desc = para_malloc(sizeof(*ui->desc));
	ui->desc->num_columns = NUM_UT_COLUMNS;
	ui->desc->flags = 0;
	ui->desc->column_descriptions = user_table_cols;
	ui->desc->dir = para_strdup(conf.database_dir_arg);
	ui->desc->name = make_message("%u", (unsigned)ui->uid);
	INFO_LOG(".............................uid #%u: %u\n",
		(unsigned)num_uids, (unsigned)ui->uid);
	if (create) {
		ret = osl(osl_create_table(ui->desc));
		if (ret < 0)
			goto err;
		num_uids++;
	}
	ret = osl(osl_open_table(ui->desc, &ui->table));
	if (ret < 0)
		goto err;
	return 1;
err:
	free((char *)ui->desc->name);
	free((char *)ui->desc->dir);
	free(ui->desc);
	ui->desc->name = NULL;
	ui->desc->dir = NULL;
	ui->desc = NULL;
	ui->table = NULL;
	ui->flags = 0;
	return ret;
}

#define uid_hash_bits 8
static uint32_t uid_hash_table_size = 1 << uid_hash_bits;
#define PRIME1 0x811c9dc5
#define PRIME2 0x01000193

static void create_hash_table(void)
{
	uid_hash_table = para_calloc(uid_hash_table_size
		* sizeof(struct user_info));
}

static void free_hash_table(void)
{
	free(uid_hash_table);
	uid_hash_table = NULL;
}

static int create_tables(void)
{
	int ret;

	dir_table_desc.dir = para_strdup(conf.database_dir_arg);
	ret = osl(osl_create_table(&dir_table_desc));
	if (ret < 0)
		return ret;
	create_hash_table();
	return 1;
}

/*
 * We use a hash table of size s=2^uid_hash_bits to map the uids into the
 * interval [0..s]. Hash collisions are treated by open addressing, i.e.
 * unused slots in the table are used to store different uids that hash to the
 * same slot.
 *
 * If a hash collision occurs, different slots are successively probed in order
 * to find an unused slot for the new uid. Probing is implemented via a second
 * hash function that maps the uid to h=(uid * PRIME2) | 1, which is always an
 * odd number.
 *
 * An odd number is sufficient to make sure each entry of the hash table gets
 * probed for probe_num between 0 and s-1 because s is a power of two, hence
 * the second hash value has never a common divisor with the hash table size.
 * IOW: h is invertible in the ring [0..s].
 */
static uint32_t double_hash(uint32_t uid, uint32_t probe_num)
{
	return (uid * PRIME1 + ((uid * PRIME2) | 1) * probe_num)
		% uid_hash_table_size;
}

#define FOR_EACH_USER(ui) for (ui = uid_hash_table; ui && ui < uid_hash_table \
		+ uid_hash_table_size; ui++)

enum search_uid_flags {
	OPEN_USER_TABLE = 1,
	CREATE_USER_TABLE = 2,
};

static int uid_is_admissible(uint32_t uid)
{
	int i;

	for (i = 0; i < conf.uid_given; i++) {
		struct uid_range *ur = admissible_uids + i;

		if (ur->low <= uid && ur->high >= uid)
			break;
	}
	i = !conf.uid_given || i < conf.uid_given;
	DEBUG_LOG("uid %u is %sadmissible\n", (unsigned)uid,
		i? "" : "not ");
	return i;
}

static int search_uid(uint32_t uid, enum search_uid_flags flags,
		struct user_info **ui_ptr)
{
	uint32_t p;

	for (p = 0; p < uid_hash_table_size; p++) {
		struct user_info *ui = uid_hash_table + double_hash(uid, p);

		if (!ui_used(ui)) {
			int ret;
			if (!flags)
				return -E_BAD_UID;
			ui->uid = uid;
			ui->flags |= UI_FL_SLOT_USED;
			if (!uid_is_admissible(uid))
				return 0;
			ui->flags |= UI_FL_ADMISSIBLE;
			ret = open_user_table(ui, flags & CREATE_USER_TABLE);
			if (ret < 0)
				return ret;

			if (ui_ptr)
				*ui_ptr = ui;
			return 1;
		}
		if (ui->uid != uid)
			continue;
		if (ui_ptr)
			*ui_ptr = ui;
		return 0;
	}
	return flags? -E_HASH_TABLE_OVERFLOW : -E_BAD_UID;
}

static int update_user_row(struct osl_table *t, uint64_t dir_num,
		uint64_t *add)
{
	struct osl_row *row;
	struct osl_object obj = {.data = &dir_num, .size = sizeof(dir_num)};

	int ret = osl(osl_get_row(t, UT_DIR_NUM, &obj, &row));

	if (ret == -E_OSL && osl_errno != E_OSL_RB_KEY_NOT_FOUND)
		return ret;
	if (ret < 0) { /* this is the first file we add */
		struct osl_object objects[NUM_UT_COLUMNS];
		uint64_t num_files = 1;

		objects[UT_DIR_NUM].data = &dir_num;
		objects[UT_DIR_NUM].size = sizeof(dir_num);
		objects[UT_BYTES].data = add;
		objects[UT_BYTES].size = sizeof(*add);
		objects[UT_FILES].data = &num_files;
		objects[UT_FILES].size = sizeof(num_files);
		INFO_LOG("######################### ret: %d\n", ret);
		ret = osl(osl_add_row(t, objects));
		INFO_LOG("######################### ret: %d\n", ret);
		return ret;
	} else { /* add size and increment file count */
		uint64_t num;
		struct osl_object obj1, obj2 = {.data = &num, .size = sizeof(num)};

		ret = osl(osl_get_object(t, row, UT_BYTES, &obj1));
		if (ret < 0)
			return ret;
		num = *(uint64_t *)obj1.data + *add;
		ret = osl(osl_update_object(t, row, UT_BYTES, &obj2));
		if (ret < 0)
			return ret;
		ret = osl(osl_get_object(t, row, UT_FILES, &obj1));
		if (ret < 0)
			return ret;
		num = *(uint64_t *)obj1.data + 1;
		return osl(osl_update_object(t, row, UT_FILES, &obj2));
	}
}

static uint64_t num_dirs;
static uint64_t num_files;
static uint64_t num_bytes;

int scan_dir(char *dirname, uint64_t *parent_dir_num)
{
	DIR *dir;
	struct dirent *entry;
	int ret, cwd_fd, ret2;
	uint64_t dir_size = 0, dir_files = 0;
	uint64_t this_dir_num = ++num_dirs;

	DEBUG_LOG("----------------- %llu: %s\n", (long long unsigned)num_dirs, dirname);
	ret = para_opendir(dirname, &dir, &cwd_fd);
	if (ret < 0) {
		if (ret != -ERRNO_TO_ERROR(EACCES))
			return ret;
		WARNING_LOG("permission denied for %s\n", dirname);
		return 1;
	}
	while ((entry = readdir(dir))) {
		mode_t m;
		struct stat s;
		uint32_t uid;
		uint64_t size;
		struct user_info *ui;

		if (!strcmp(entry->d_name, "."))
			continue;
		if (!strcmp(entry->d_name, ".."))
			continue;
		if (lstat(entry->d_name, &s) == -1) {
			WARNING_LOG("lstat error for %s/%s\n", dirname,
				entry->d_name);
			continue;
		}
		m = s.st_mode;
		if (!S_ISREG(m) && !S_ISDIR(m))
			continue;
		if (S_ISDIR(m)) {
			ret = scan_dir(entry->d_name, &this_dir_num);
			if (ret < 0)
				goto out;
			continue;
		}
		/* regular file */
		size = s.st_size;
		dir_size += size;
		num_bytes += size;
		dir_files++;
		num_files++;
		uid = s.st_uid;
		ret = search_uid(uid, CREATE_USER_TABLE | OPEN_USER_TABLE, &ui);
		if (ret < 0)
			goto out;
		ui->bytes += size;
		ui->files++;
		ret = update_user_row(ui->table, this_dir_num, &size);
		if (ret < 0)
			goto out;
	}
	ret = add_directory(dirname, &this_dir_num, parent_dir_num,
			&dir_size, &dir_files);
out:
	closedir(dir);
	ret2 = para_fchdir(cwd_fd);
	if (ret2 < 0 && ret >= 0)
		ret = ret2;
	close(cwd_fd);
	return ret;
}

static int get_dir_name_by_number(uint64_t *dirnum, char **name)
{
	char *result = NULL, *tmp;
	struct osl_row *row;
	uint64_t val = *dirnum;
	struct osl_object obj = {.data = &val, .size = sizeof(val)};
	int ret;

again:
	ret = osl(osl_get_row(dir_table, DT_NUM, &obj, &row));
	if (ret < 0)
		goto out;
	ret = osl(osl_get_object(dir_table, row, DT_NAME, &obj));
	if (ret < 0)
		goto out;
	if (result) {
		tmp = make_message("%s/%s", (char *)obj.data, result);
		free(result);
		result = tmp;
	} else
		result = para_strdup((char *)obj.data);
	ret = osl(osl_get_object(dir_table, row, DT_PARENT_NUM, &obj));
	if (ret < 0)
		goto out;
	val = *(uint64_t *)obj.data;
	if (val)
		goto again;
out:
	if (ret < 0) {
		free(result);
		*name = NULL;
	} else
		*name = result;
	return ret;
}

static int get_dir_name_of_row(struct osl_row *dir_table_row, char **name)
{
	struct osl_object obj;
	int ret;
	char *this_dir, *prefix = NULL;

	*name = NULL;
	ret = osl(osl_get_object(dir_table, dir_table_row, DT_NAME, &obj));
	if (ret < 0)
		return ret;
	this_dir = para_strdup((char *)obj.data);
	ret = osl(osl_get_object(dir_table, dir_table_row, DT_PARENT_NUM, &obj));
	if (ret < 0)
		goto out;
	if (!*(uint64_t *)obj.data) {
		*name = this_dir;
		return 1;
	}
	ret = get_dir_name_by_number((uint64_t *)obj.data, &prefix);
	if (ret < 0)
		goto out;
	*name = make_message("%s/%s", prefix, this_dir);
	free(prefix);
	ret = 1;
out:
	free(this_dir);
	return ret;
}

const uint64_t size_unit_divisors[] = {
	[size_unit_arg_b] = 1ULL,
	[size_unit_arg_k] = 1024ULL,
	[size_unit_arg_m] = 1024ULL * 1024ULL,
	[size_unit_arg_g] = 1024ULL * 1024ULL * 1024ULL,
	[size_unit_arg_t] = 1024ULL * 1024ULL * 1024ULL * 1024ULL,
};

const uint64_t count_unit_divisors[] = {

	[count_unit_arg_n] = 1ULL,
	[count_unit_arg_k] = 1000ULL,
	[count_unit_arg_m] = 1000ULL * 1000ULL,
	[count_unit_arg_g] = 1000ULL * 1000ULL * 1000ULL,
	[count_unit_arg_t] = 1000ULL * 1000ULL * 1000ULL * 1000ULL,
};

const char size_unit_abbrevs[] = " BKMGT";
const char count_unit_abbrevs[] = "  KMGT";

static void format_size_value(enum enum_size_unit unit, uint64_t value, char *result)
{
	if (unit == size_unit_arg_h) /* human readable */
		for (unit = size_unit_arg_b; unit < size_unit_arg_t && value > size_unit_divisors[unit + 1]; unit++)
				; /* nothing */
	sprintf(result, "%llu%c", (long long unsigned)value / size_unit_divisors[unit], size_unit_abbrevs[unit]);
}

static void format_count_value(enum enum_count_unit unit, uint64_t value, char *result)
{
	if (unit == count_unit_arg_h) /* human readable */
		for (unit = count_unit_arg_n; unit < count_unit_arg_t && value > count_unit_divisors[unit + 1]; unit++)
				; /* nothing */
	sprintf(result, "%llu%c", (long long unsigned)value / count_unit_divisors[unit], count_unit_abbrevs[unit]);
}

enum global_stats_flags {
	GSF_PRINT_DIRNAME = 1,
	GSF_PRINT_BYTES = 2,
	GSF_PRINT_FILES = 4,
	GSF_COMPUTE_SUMMARY = 8,
};

struct global_stats_info {
	uint32_t count;
	int ret;
	int osl_errno;
	enum global_stats_flags flags;
};

static int global_stats_loop_function(struct osl_row *row, void *data)
{
	struct global_stats_info *gsi = data;
	struct osl_object obj;
	char *dirname, formated_value[25];
	int ret, summary = gsi->flags & GSF_COMPUTE_SUMMARY;

	if (!gsi->count && !summary) {
		ret = -E_LOOP_COMPLETE;
		goto err;
	}
	if (gsi->count && (gsi->flags & GSF_PRINT_DIRNAME)) {
		ret = get_dir_name_of_row(row, &dirname);
		if (ret < 0)
			goto err;
		printf("%s%s", dirname,
			(gsi->flags & (GSF_PRINT_FILES | GSF_PRINT_BYTES))?
				"\t" : "\n"
		);
	}
	if (summary || (gsi->count && (gsi->flags & GSF_PRINT_FILES))) {
		uint64_t files;
		ret = osl(osl_get_object(dir_table, row, DT_FILES, &obj));
		if (ret < 0)
			goto err;
		files = *(uint64_t *)obj.data;
		if (gsi->count && (gsi->flags & GSF_PRINT_FILES)) {
			format_size_value(conf.size_unit_arg, files,
				formated_value);
			printf("%s%s", formated_value,
				(gsi->flags & GSF_PRINT_BYTES)? "\t" : "\n");
		}
		if (summary)
			num_files += files;
	}
	if (summary || (gsi->count && (gsi->flags & GSF_PRINT_BYTES))) {
		uint64_t bytes;
		ret = osl(osl_get_object(dir_table, row, DT_BYTES, &obj));
		if (ret < 0)
			goto err;
		bytes = *(uint64_t *)obj.data;
		if (gsi->count && (gsi->flags & GSF_PRINT_BYTES)) {
			format_size_value(conf.size_unit_arg, bytes,
				formated_value);
			printf("%s\n", formated_value);
		}
		if (summary) {
			num_bytes += bytes;
			num_dirs++;
		}
	}
	if (gsi->count > 0)
		gsi->count--;
	return 1;
err:
	gsi->ret = ret;
	gsi->osl_errno = (ret == -E_OSL)? osl_errno : 0;
	return -1;
}

static void print_id_stats(void)
{
	struct user_info *ui;

	printf("--------------------- user summary (uid/dirs/files/bytes):\n");
	FOR_EACH_USER(ui) {
		char formated_dir_count[25], formated_file_count[25],
			formated_bytes[25];
		if (!ui_used(ui) || !ui_admissible(ui))
			continue;
		format_count_value(conf.count_unit_arg, ui->dirs,
			formated_dir_count);
		format_count_value(conf.count_unit_arg, ui->files,
			formated_file_count);
		format_size_value(conf.size_unit_arg, ui->bytes,
			formated_bytes);
		printf("%u\t%s\t%s\t%s\n", (unsigned)ui->uid,
			formated_dir_count,
			formated_file_count,
			formated_bytes
		);
	}
}

enum user_stats_flags {
	USF_PRINT_DIRNAME = 1,
	USF_PRINT_BYTES = 2,
	USF_PRINT_FILES = 4,
	USF_COMPUTE_SUMMARY = 8,
};

struct user_stats_info {
	uint32_t count;
	enum user_stats_flags flags;
	int ret;
	int osl_errno;
	struct user_info *ui;
};

static int user_stats_loop_function(struct osl_row *row, void *data)
{
	struct user_stats_info *usi = data;
	struct osl_object obj;
	int ret, summary = usi->flags & GSF_COMPUTE_SUMMARY;
	char formated_value[25];

	if (!usi->count && !summary) {
		ret = -E_LOOP_COMPLETE;
		goto err;
	}
	if (usi->count && (usi->flags & USF_PRINT_DIRNAME)) {
		char *dirname;
		ret = osl(osl_get_object(usi->ui->table, row, UT_DIR_NUM, &obj));
		if (ret < 0)
			goto err;
		ret = get_dir_name_by_number((uint64_t *)obj.data, &dirname);
		if (ret < 0)
			goto err;
		printf("%s%s",
			dirname,
			(usi->flags & (USF_PRINT_FILES | USF_PRINT_BYTES))?
				"\t" : "\n"
		);
	}
	if (summary || (usi->count && (usi->flags & USF_PRINT_FILES))) {
		uint64_t files;
		ret = osl(osl_get_object(usi->ui->table, row, UT_FILES, &obj));
		if (ret < 0)
			goto err;
		files = *(uint64_t *)obj.data;
		if (usi->count && (usi->flags & USF_PRINT_FILES)) {
			format_size_value(conf.size_unit_arg, files,
				formated_value);
			printf("%s%s", formated_value,
				(usi->flags & USF_PRINT_BYTES)? "\t" : "\n"
			);
		}
		if (summary)
			usi->ui->files += files;
	}
	if (summary || (usi->count && (usi->flags & USF_PRINT_BYTES))) {
		uint64_t bytes;
		ret = osl(osl_get_object(usi->ui->table, row, UT_BYTES, &obj));
		if (ret < 0)
			goto err;
		bytes = *(uint64_t *)obj.data;
		if (usi->count && (usi->flags & USF_PRINT_BYTES)) {
			format_size_value(conf.size_unit_arg, bytes,
				formated_value);
			printf("%s\n", formated_value);
		}
		if (summary) {
			usi->ui->bytes += bytes;
			usi->ui->dirs++;
		}

	}
	if (usi->count > 0)
		usi->count--;
	return 1;
err:
	usi->ret = ret;
	usi->osl_errno = (ret == -E_OSL)? osl_errno : 0;
	return -1;
}

static int check_loop_return(int ret, int loop_ret, int loop_osl_errno)
{
	if (ret >= 0)
		return ret;
	assert(ret == -E_OSL);
	if (osl_errno != E_OSL_LOOP)
		/* error not caused by loop function returning negative. */
		return ret;
	assert(loop_ret < 0);
	if (loop_ret == -E_LOOP_COMPLETE) /* no error */
		return 1;
	if (loop_ret == -E_OSL) { /* osl error in loop function */
		assert(loop_osl_errno);
		osl_errno = loop_osl_errno;
	}
	return loop_ret;
}

static int adu_loop_reverse(struct osl_table *t, unsigned col_num, void *private_data,
		osl_rbtree_loop_func *func, int *loop_ret, int *loop_osl_errno)
{
	int ret = osl(osl_rbtree_loop_reverse(t, col_num, private_data, func));
	return check_loop_return(ret, *loop_ret, *loop_osl_errno);
}

static int print_user_stats(void)
{
	struct user_info *ui;
	int ret;

	FOR_EACH_USER(ui) {
		struct user_stats_info usi = {
			.count = conf.limit_arg,
			.ui = ui
		};
		if (!ui_used(ui) || !ui_admissible(ui))
			continue;
		usi.flags = USF_PRINT_DIRNAME | USF_PRINT_BYTES | USF_COMPUTE_SUMMARY;
		printf("************************************************ uid %u\n",
			(unsigned) ui->uid);
		printf("----------------- Largest dirs -------------------\n");
		ret = adu_loop_reverse(ui->table, UT_BYTES, &usi, user_stats_loop_function,
			&usi.ret, &usi.osl_errno);
		if (ret < 0)
			return ret;
		printf("---------- dirs containing most files ------------\n");
		usi.count = conf.limit_arg,
		usi.flags = USF_PRINT_DIRNAME | USF_PRINT_FILES;
		ret = adu_loop_reverse(ui->table, UT_FILES, &usi, user_stats_loop_function,
			&usi.ret, &usi.osl_errno);
		if (ret < 0)
			return ret;
	}
	return 1;
}

static int print_statistics(void)
{
	int ret;
	struct global_stats_info gsi = {
		.count = conf.limit_arg,
		.flags = GSF_PRINT_DIRNAME | GSF_PRINT_BYTES | GSF_COMPUTE_SUMMARY
	};

	printf("----------------- Largest dirs -------------------\n");
	ret = adu_loop_reverse(dir_table, DT_BYTES, &gsi,
		global_stats_loop_function, &gsi.ret, &gsi.osl_errno);
	if (ret < 0)
		return ret;
	gsi.count = conf.limit_arg;

	gsi.flags = GSF_PRINT_DIRNAME | GSF_PRINT_FILES;
	printf("---------- dirs containing most files ------------\n");
	ret = adu_loop_reverse(dir_table, DT_FILES, &gsi,
		global_stats_loop_function, &gsi.ret, &gsi.osl_errno);
	if (ret < 0)
		return ret;
	printf("------------------ Global summary (dirs/files/bytes)\n"
		"%llu\t%llu\t%llu\n",
		(long long unsigned)num_dirs, (long long unsigned)num_files,
		(long long unsigned)num_bytes);
	print_user_stats();
	print_id_stats();
	return 1;
}

static char *get_uid_list_name(void)
{
	return make_message("%s/uid_list", conf.database_dir_arg);
}

static int write_uid_list(void)
{
	char *buf, *filename;
	uint32_t count = 0;
	struct user_info *ui;
	size_t size = num_uids * sizeof(uint32_t);
	int ret;

	if (!num_uids)
		return 0;
	buf = para_malloc(size);
	FOR_EACH_USER(ui) {
		if (!ui_used(ui) || !ui_admissible(ui))
			continue;
		DEBUG_LOG("saving uid %u\n", (unsigned) ui->uid);
		write_u32(buf + count++ * sizeof(uint32_t), ui->uid);
	}
	filename = get_uid_list_name();
	ret = para_write_file(filename, buf, size);
	free(filename);
	free(buf);
	return ret;
}

static int open_dir_table(void)
{
	if (!dir_table_desc.dir) /* we did not create the table */
		dir_table_desc.dir = para_strdup(conf.database_dir_arg);
	return osl(osl_open_table(&dir_table_desc, &dir_table));
}

static void close_dir_table(void)
{
	int ret;

	if (!dir_table)
		return;
	ret = osl(osl_close_table(dir_table, OSL_MARK_CLEAN));
	if (ret < 0)
		ERROR_LOG("failed to close dir table: %s\n", adu_strerror(-ret));
	free((char *)dir_table_desc.dir);
	dir_table = NULL;
}

static void close_user_table(struct user_info *ui)
{
	int ret;

	if (!ui || !ui_used(ui) || !ui_admissible(ui))
		return;
	ret = osl(osl_close_table(ui->table, OSL_MARK_CLEAN));
	if (ret < 0)
		ERROR_LOG("failed to close user table %u: %s\n",
			(unsigned) ui->uid, adu_strerror(-ret));
	free((char *)ui->desc->name);
	ui->desc->name = NULL;
	free((char *)ui->desc->dir);
	ui->desc->dir = NULL;
	free(ui->desc);
	ui->desc = NULL;
	ui->table = NULL;
	ui->flags = 0;
}

static void close_user_tables(void)
{
	struct user_info *ui;

	FOR_EACH_USER(ui)
		close_user_table(ui);
}

static void close_all_tables(void)
{
	close_dir_table();
	close_user_tables();
	free_hash_table();
}

static int com_create()
{
	uint64_t zero = 0ULL;
	int ret = create_tables();

	if (ret < 0)
		return ret;
	ret = open_dir_table();
	if (ret < 0)
		return ret;
	ret = scan_dir(conf.base_dir_arg, &zero);
	if (ret < 0)
		goto out;
	ret = write_uid_list();
out:
	close_all_tables();
	return ret;
}

static int read_uid_file(void)
{
	size_t size;
	uint32_t n;
	char *filename = get_uid_list_name(), *map;
	int ret = mmap_full_file(filename, O_RDONLY, (void **)&map, &size, NULL);

	if (ret < 0) {
		INFO_LOG("failed to map %s\n", filename);
		free(filename);
		return ret;
	}
	num_uids = size / 4;
	INFO_LOG("found %u uids in %s\n", (unsigned)num_uids, filename);
	free(filename);
	/* hash table size should be a power of two and larger than the number of uids */
	uid_hash_table_size = 4;
	while (uid_hash_table_size < num_uids)
		uid_hash_table_size *= 2;
	create_hash_table();
	for (n = 0; n < num_uids; n++) {
		uint32_t uid = read_u32(map + n * sizeof(uid));
		ret = search_uid(uid, OPEN_USER_TABLE, NULL);
		if (ret < 0)
			goto out;
	}
out:
	para_munmap(map, size);
	return ret;
}

static int com_select(void)
{
	int ret;

	ret = open_dir_table();
	if (ret < 0)
		return ret;
	ret = read_uid_file();
	if (ret < 0)
		return ret;
	ret = print_statistics();
	close_all_tables();
	return ret;
}

static int check_args(void)
{
	int i, ret;

	/* remove trailing slashes from base-dir arg */
	if (conf.base_dir_given) {
		size_t len = strlen(conf.base_dir_arg);
		for (;;) {
			if (!len) /* empty string */
				return -ERRNO_TO_ERROR(EINVAL);
			if (!--len) /* length 1 is always OK */
				break;
			if (conf.base_dir_arg[len] != '/')
				break; /* no trailing slash, also OK */
			conf.base_dir_arg[len] = '\0';
		}
	}
	if (!conf.uid_given)
		return 0;
	admissible_uids = para_malloc(conf.uid_given * sizeof(*admissible_uids));
	for (i = 0; i < conf.uid_given; i++) {
		ret = parse_uid_range(conf.uid_arg[i], admissible_uids + i);
		if (ret < 0)
			goto err;
	}
	return 1;
err:
	free(admissible_uids);
	admissible_uids = NULL;
	return ret;
}

int main(int argc, char **argv)
{
	int ret;
	struct cmdline_parser_params params = {
		.override = 0,
		.initialize = 1,
		.check_required = 0,
		.check_ambiguity = 0,
		.print_errors = 1
	};

	cmdline_parser_ext(argc, argv, &conf, &params); /* aborts on errors */
	ret = check_args();
	if (ret < 0)
		goto out;
	ret = -E_SYNTAX;
	if (conf.select_given)
		ret = com_select();
	else
		ret = com_create();
	if (ret < 0)
		goto out;
out:
	free(admissible_uids);
	if (ret < 0) {
		ERROR_LOG("%s\n", adu_strerror(-ret));
		return -EXIT_FAILURE;
	}
	return EXIT_SUCCESS;
}
