/*
 * Copyright (C) 2006-2008 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file fd.c Helper functions for file descriptor handling. */

#include <sys/types.h>
#include <dirent.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <sys/select.h>

#include "adu.h"
#include "error.h"

/**
 * Wrapper for the write system call.
 *
 * \param fd The file descriptor to write to.
 * \param buf The buffer to write.
 * \param size The length of \a buf in bytes.
 *
 * This function writes out the given buffer and retries if an interrupt
 * occurred during the write.
 *
 * \return On success, the number of bytes written is returned, otherwise, the
 * function returns \p -E_WRITE.
 *
 * \sa write(2).
 */
ssize_t para_write(int fd, const void *buf, size_t size)
{
	ssize_t ret;

	for (;;) {
		ret = write(fd, buf, size);
		if ((ret < 0) && (errno == EAGAIN || errno == EINTR))
			continue;
		return ret >= 0? ret : -ERRNO_TO_ERROR(errno);
	}
}


/**
 * Write the whole buffer to a file descriptor.
 *
 * \param fd The file descriptor to write to.
 * \param buf The buffer to write.
 * \param size The length of \a buf in bytes.
 *
 * This function writes the given buffer and continues on short writes and
 * when interrupted by a signal.
 *
 * \return Positive on success, negative on errors. Possible errors: any
 * errors returned by para_write().
 *
 * \sa para_write().
 */
ssize_t write_all(int fd, const void *buf, size_t size)
{
//	DEBUG_LOG("writing %zu bytes\n", size);
	const char *b = buf;
	while (size) {
		ssize_t ret = para_write(fd, b, size);
//		DEBUG_LOG("ret: %zd\n", ret);
		if (ret < 0)
			return ret;
		b += ret;
		size -= ret;
	}
	return 1;
}

/**
 * Wrapper for the open(2) system call.
 *
 * \param path The filename.
 * \param flags The usual open(2) flags.
 * \param mode Specifies the permissions to use.
 *
 * The mode parameter must be specified when O_CREAT is in the flags, and is
 * ignored otherwise.
 *
 * \return The file descriptor on success, negative on errors.
 *
 * \sa open(2).
 */
int para_open(const char *path, int flags, mode_t mode)
{
	int ret = open(path, flags, mode);

	if (ret >= 0)
		return ret;
	return -ERRNO_TO_ERROR(errno);
}

/**
 * Open a file, write the given buffer and close the file.
 *
 * \param filename Full path to the file to open.
 * \param buf The buffer to write to the file.
 * \param size The size of \a buf.
 *
 * \return Positive on success, negative on errors. Possible errors include:
 * any errors from para_open() or para_write().
 *
 * \sa para_open(), para_write().
 */
int para_write_file(const char *filename, const void *buf, size_t size)
{
	int ret, fd;

	ret = para_open(filename, O_WRONLY | O_CREAT | O_EXCL, 0644);
	if (ret < 0)
		return ret;
	fd = ret;
	ret = write_all(fd, buf, size);
	if (ret < 0)
		goto out;
	ret = 1;
out:
	close(fd);
	return ret;
}


/**
 * Check whether a file exists.
 *
 * \param fn The file name.
 *
 * \return Non-zero iff file exists.
 */
int file_exists(const char *fn)
{
	struct stat statbuf;

	return !stat(fn, &statbuf);
}

/**
 * Paraslash's wrapper for select(2).
 *
 * It calls select(2) (with no exceptfds) and starts over if select() was
 * interrupted by a signal.
 *
 * \param n The highest-numbered descriptor in any of the two sets, plus 1.
 * \param readfds fds that should be checked for readability.
 * \param writefds fds that should be checked for writablility.
 * \param timeout_tv upper bound on the amount of time elapsed before select()
 * returns.
 *
 * \return The return value of the underlying select() call on success, the
 * negative system error code on errors.
 *
 * All arguments are passed verbatim to select(2).
 * \sa select(2) select_tut(2).
 */
int para_select(int n, fd_set *readfds, fd_set *writefds,
		struct timeval *timeout_tv)
{
	int ret, err;
	do {
		ret = select(n, readfds, writefds, NULL, timeout_tv);
		err = errno;
	} while (ret < 0 && err == EINTR);
	if (ret < 0)
		return -ERRNO_TO_ERROR(errno);
	return ret;
}

/**
 * Set a file descriptor to blocking mode.
 *
 * \param fd The file descriptor.
 *
 * \return Standard.
 */
__must_check int mark_fd_blocking(int fd)
{
	int flags = fcntl(fd, F_GETFL);
	if (flags < 0)
		return -ERRNO_TO_ERROR(errno);
	flags = fcntl(fd, F_SETFL, ((long)flags) & ~O_NONBLOCK);
	if (flags < 0)
		return -ERRNO_TO_ERROR(errno);
	return 1;
}

/**
 * Set a file descriptor to non-blocking mode.
 *
 * \param fd The file descriptor.
 *
 * \return Standard.
 */
__must_check int mark_fd_nonblocking(int fd)
{
	int flags = fcntl(fd, F_GETFL);
	if (flags < 0)
		return -ERRNO_TO_ERROR(errno);
	flags = fcntl(fd, F_SETFL, ((long)flags) | O_NONBLOCK);
	if (flags < 0)
		return -ERRNO_TO_ERROR(errno);
	return 1;
}

/**
 * Set a file descriptor in a fd_set.
 *
 * \param fd The file descriptor to be set.
 * \param fds The file descriptor set.
 * \param max_fileno Highest-numbered file descriptor.
 *
 * This wrapper for FD_SET() passes its first two arguments to \p FD_SET. Upon
 * return, \a max_fileno contains the maximum of the old_value and \a fd.
 *
 * \sa para_select.
*/
void para_fd_set(int fd, fd_set *fds, int *max_fileno)
{

	if (fd < 0 || fd >= FD_SETSIZE) {
		EMERG_LOG("fatal: tried to add invalid fd %d\n", fd);
		exit(EXIT_FAILURE);
	}
#if 0
	{
		int flags = fcntl(fd, F_GETFL);
		if (!(flags & O_NONBLOCK)) {
			EMERG_LOG("fd %d is a blocking file descriptor\n", fd);
			exit(EXIT_FAILURE);
		}
	}
#endif
	FD_SET(fd, fds);
	*max_fileno = MAX(*max_fileno, fd);
}

/**
 * Paraslash's wrapper for mmap.
 *
 * \param length Number of bytes to mmap.
 * \param prot Either PROT_NONE or the bitwise OR of one or more of
 * PROT_EXEC PROT_READ PROT_WRITE.
 * \param flags Exactly one of MAP_SHARED and MAP_PRIVATE.
 * \param fd The file to mmap from.
 * \param offset Mmap start.
 *
 * \return This function either returns a valid pointer to the mapped area
 * or calls exit() on errors.
 */
void *para_mmap(size_t length, int prot, int flags, int fd, off_t offset)
{
	void *ret = mmap(NULL, length, prot, flags, fd, offset);
	if (ret != MAP_FAILED)
		return ret;
	EMERG_LOG("mmap failed: %s\n", strerror(errno));
	EMERG_LOG("length: %zu, flags: %d, fd: %d, offset: %zu\n",
		length, flags, fd, (size_t)offset);
	exit(EXIT_FAILURE);
}

/**
 * Wrapper for chdir(2).
 *
 * \param path The specified directory.
 *
 * \return Standard.
 */
int para_chdir(const char *path)
{
	int ret = chdir(path);

	if (ret >= 0)
		return 1;
	return -ERRNO_TO_ERROR(errno);
}

/**
 * Save the cwd and open a given directory.
 *
 * \param dirname Path to the directory to open.
 * \param dir Result pointer.
 * \param cwd File descriptor of the current working directory.
 *
 * \return Standard.
 *
 * Opening the current directory (".") and calling fchdir() to return is
 * usually faster and more reliable than saving cwd in some buffer and calling
 * chdir() afterwards.
 *
 * If \a cwd is not \p NULL "." is opened and the resulting file descriptor is
 * stored in \a cwd. If the function returns success, and \a cwd is not \p
 * NULL, the caller must close this file descriptor (probably after calling
 * fchdir(*cwd)).
 *
 * On errors, the function undos everything, so the caller needs neither close
 * any files, nor change back to the original working directory.
 *
 * \sa getcwd(3).
 *
 */
int para_opendir(const char *dirname, DIR **dir, int *cwd)
{
	int ret;

	if (cwd) {
		ret = para_open(".", O_RDONLY, 0);
		if (ret < 0)
			return ret;
		*cwd = ret;
	}
	ret = para_chdir(dirname);
	if (ret < 0)
		goto close_cwd;
	*dir = opendir(".");
	if (*dir)
		return 1;
	ret = -ERRNO_TO_ERROR(errno);
/* Ignore return value of fchdir() and close(). We're busted anyway. */
	if (cwd)
		fchdir(*cwd);
close_cwd:
	if (cwd)
		close(*cwd);
	return ret;
}

/**
 * A wrapper for fchdir().
 *
 * \param fd An open file descriptor.
 *
 * \return Standard.
 */
int para_fchdir(int fd)
{
	if (fchdir(fd) < 0)
		return -ERRNO_TO_ERROR(errno);
	return 1;
}

/**
 * A wrapper for mkdir(2).
 *
 * \param path Name of the directory to create.
 * \param mode The permissions to use.
 *
 * \return Standard.
 */
int para_mkdir(const char *path, mode_t mode)
{
	if (!mkdir(path, mode))
		return 1;
	return -ERRNO_TO_ERROR(errno);
}

/**
 * Open a file and map it into memory.
 *
 * \param path Name of the regular file to map.
 * \param open_mode Either \p O_RDONLY or \p O_RDWR.
 * \param map On success, the mapping is returned here.
 * \param size size of the mapping.
 * \param fd_ptr The file descriptor of the mapping.
 *
 * If \a fd_ptr is \p NULL, the file descriptor resulting from the underlying
 * open call is closed after mmap().  Otherwise the file is kept open and the
 * file descriptor is returned in \a fd_ptr.
 *
 * \return Standard.
 *
 * \sa para_open(), mmap(2).
 */
int mmap_full_file(const char *path, int open_mode, void **map,
		size_t *size, int *fd_ptr)
{
	int fd, ret, mmap_prot, mmap_flags;
	struct stat file_status;

	if (open_mode == O_RDONLY) {
		mmap_prot = PROT_READ;
		mmap_flags = MAP_PRIVATE;
	} else {
		mmap_prot = PROT_READ | PROT_WRITE;
		mmap_flags = MAP_SHARED;
	}
	ret = para_open(path, open_mode, 0);
	if (ret < 0)
		return ret;
	fd = ret;
	if (fstat(fd, &file_status) < 0) {
		ret = -ERRNO_TO_ERROR(errno);
		goto out;
	}
	*size = file_status.st_size;
	ret = -E_EMPTY;
	DEBUG_LOG("%s: size %zu\n", path, *size);
	if (!*size)
		goto out;
	*map = mmap(NULL, *size, mmap_prot, mmap_flags, fd, 0);
	if (*map == MAP_FAILED) {
		*map = NULL;
		ret = -E_MMAP;
		goto out;
	}
	ret = 1;
out:
	if (ret < 0 || !fd_ptr)
		close(fd);
	else
		*fd_ptr = fd;
	return ret;
}

/**
 * A wrapper for munmap(2).
 *
 * \param start The start address of the memory mapping.
 * \param length The size of the mapping.
 *
 * \return Positive on success, \p -E_MUNMAP on errors.
 *
 * \sa munmap(2), mmap_full_file().
 */
int para_munmap(void *start, size_t length)
{
	int err;
	if (munmap(start, length) >= 0)
		return 1;
	err = errno;
	ERROR_LOG("munmap (%p/%zu) failed: %s\n", start, length,
		strerror(err));
	return -ERRNO_TO_ERROR(err);
}

/**
 * Check a file descriptor for writability.
 *
 * \param fd The file descriptor.
 *
 * \return positive if fd is ready for writing, zero if it isn't, negative if
 * an error occurred.
 */

int write_ok(int fd)
{
	struct timeval tv = {0, 0};
	fd_set wfds;
	int ret;
again:
	FD_ZERO(&wfds);
	FD_SET(fd, &wfds);
	tv.tv_sec = 0;
	tv.tv_usec = 0;
	ret = select(fd + 1, NULL, &wfds, NULL, &tv);
	if (ret < 0 && errno == EINTR)
		goto again;
	return ret;
}
